IMPLEMENTATION MODULE FormAusLib;

  (*------------------------------------------------------------------------*)
  (* Routinen die gemeinsam von {T|F|S}FormAus genutzt werden               *)
  (*------------------------------------------------------------------------*)
  (* Letzte Bearbeitung:                                                    *)
  (*                                                                        *)
  (* 05.09.94, MRi: Durchsicht                                              *)
  (* 22.06.15, MRi: Einf"uhren des "An"-Formats in WriteN,dWriteN           *)
  (* 02.10.16, MRi: Aufsplitten der Module FormAus und SFormAus in          *)
  (*                TFormAus,FFormAus,SFormAus                              *)
  (* 02.12.16, MRi: In ByteToLngReal Umwandlung von 4-BYTE Reals eingefuegt *)
  (*------------------------------------------------------------------------*)
  (* Offene Punkte                                                          *)
  (*                                                                        *)
  (* - Eventuell die Formatanweisung ausgeben wenn eine ByteToXXX Routinen  *)
  (*   einen Fehler meldet so dass diese leichter gefunden werden kann.     *)
  (* - Durchsicht der Laengenparameter fuer Cardinal                        *)
  (* - Ausschreiben einer LONGINT wie LONGCARD behandeln                    *)
  (* - Ausschreiben von REAL einfuegen - Mechanismus wie bei CARDINAL       *)
  (*------------------------------------------------------------------------*)
  (* Licence : GNU Lesser General Public License (LGPL)                     *)
  (*------------------------------------------------------------------------*)

  (* $Id: FormAusLib.mod,v 1.2 2018/01/05 13:41:13 mriedl Exp mriedl $ *)

FROM SYSTEM     IMPORT BYTE,TSIZE,ADR;
FROM Deklera    IMPORT STRING;
FROM Errors     IMPORT Fehler,ErrOut;
                IMPORT TIO;


PROCEDURE WrByteStr(Str    : ARRAY OF BYTE; (* F"ur Mocka *)
                    Laenge : CARDINAL);

          VAR String : StrPrt;
BEGIN
      String:=ADR(Str);
      IF (Laenge > TSIZE(STRING)) THEN 
        ErrOut("Zeichenkette zu lang (TFormAus.ByteToStr) !");
        Laenge:=TSIZE(STRING) - 1; 
      END;
      String^[Laenge]:=0C;
      TIO.WrStr(String^);
END WrByteStr;

PROCEDURE WrByteStrN(Str    : ARRAY OF BYTE; (* F"ur Mocka *)
                     Laenge : CARDINAL);

          VAR String : StrPrt;
BEGIN
      String:=ADR(Str);
      IF (Laenge > TSIZE(STRING)) THEN 
        ErrOut("Zeichenkette zu lang (TFormAus.ByteToStr) !");
        Laenge:=TSIZE(STRING) - 1; 
      END;
      String^[Laenge]:=0C;
      TIO.WrStrForm(String^,-VAL(INTEGER,Laenge));
END WrByteStrN;

PROCEDURE ByteToStr(Str    : ARRAY OF BYTE;
                    Laenge : CARDINAL) : StrPrt;

          VAR String : StrPrt;
BEGIN
      String:=ADR(Str);
      IF (Laenge > TSIZE(STRING)) THEN 
        ErrOut("Zeichenkette zu lang (TFormAus.ByteToStr) !");
        Laenge:=TSIZE(STRING) - 1; 
      END;
      String^[Laenge]:=0C; (* Hier ausschreiben klappt noch mit Mocka *)
      RETURN String;
END ByteToStr;

PROCEDURE ByteToInt(i      : ARRAY OF BYTE;
                    Laenge : CARDINAL) : INTEGER;

          VAR I : POINTER TO INTEGER;
BEGIN
      IF (Laenge # TSIZE(INTEGER)) THEN
        Fehler:=TRUE;
        TIO.WrStr('Fehler in ByteToInt.');
        RETURN MAX(INTEGER);
      ELSE
        I:=ADR(i); RETURN I^;
      END;
END ByteToInt;

PROCEDURE ByteToCard(i      : ARRAY OF BYTE;
                     Laenge : CARDINAL) : CARDINAL;

          VAR I : POINTER TO CARDINAL;
BEGIN
      IF (Laenge # TSIZE(CARDINAL)) THEN
        Fehler:=TRUE;
        TIO.WrStr('Fehler in ByteToCard.');
        RETURN MAX(CARDINAL);
      ELSE
        I:=ADR(i); RETURN I^;
      END;
END ByteToCard;

PROCEDURE ByteToLngCard(i      : ARRAY OF BYTE;
                        Laenge : CARDINAL) : LONGCARD;

          VAR I : POINTER TO LONGCARD;
BEGIN
      IF (Laenge # TSIZE(LONGCARD)) THEN
        Fehler:=TRUE;
        TIO.WrStr('Fehler in ByteToLngCard.');
        RETURN MAX(LONGCARD);
      ELSE
        I:=ADR(i); RETURN I^;
      END;
END ByteToLngCard;

PROCEDURE ByteToLngReal(x      : ARRAY OF BYTE;
                        Laenge : CARDINAL) : LONGREAL;

          VAR R8 : POINTER TO LONGREAL;
              R4 : POINTER TO REAL;
              r8 : LONGREAL;
BEGIN
      IF (Laenge # TSIZE(LONGREAL)) AND (Laenge # TSIZE(REAL)) THEN
        Fehler:=TRUE;
        TIO.WrStr('Fehler in ByteToLngReal.');
        RETURN MAX(LONGREAL);
      END;
      IF (Laenge = TSIZE(LONGREAL)) THEN
        R8:=ADR(x);
        RETURN R8^;
      ELSE
        R4:=ADR(x); r8:=VAL(LONGREAL,R4^);
        RETURN r8;
      END;
END ByteToLngReal;

END FormAusLib.
