IMPLEMENTATION MODULE NumConvert; 

  (*------------------------------------------------------------------------*)
  (* Modul fuer die Konvertierung Ganzahlen / Sting et. visa versa.         *)
  (* Modul for the conversion of whole numbers to string or visa versa.     *)
  (*------------------------------------------------------------------------*)
  (* Die Implementation basieren zum Teil auf verschiedenen Quellen die     *)
  (* "uberarbeitet wurden                                                   *)
  (*------------------------------------------------------------------------*)
  (* Letzte Bearbeitung:                                                    *)
  (*                                                                        *)
  (* 01.08.93, MRi: Durchsicht, Mocka / MtC u.a. fuer OS/2, Linux, Solaris  *)
  (*       95, MRi: Abspalten der Fliesskommazahl-Prozeduren                *)
  (* 09.11.14, MRi: Kleinere Anpassungen an ISO M2 (Typkonversionen)        *)
  (* 28.11.14, MRi: Kleinere Anpassungen (Verschoenerungen ;-)              *)
  (* 29.11.14, MRi: Einfuerung der Konstanten Max/MinInt,MaxCard            *)
  (* 01.12.14, MRi: Initial Version von CardToBitStr eingefuegt             *)
  (* 18.02.15, MRi: Korrektur fuer nbits=MaxBits um Ueberlauf von pot2 in   *)
  (*                CardToBitStr zu vermeiden.                              *)
  (* 10.11.15, MRi: Kopf aktualisier, GPL Notiz eingefuegt.                 *)
  (*                ConvStr jetzt Konstante                                 *)
  (*------------------------------------------------------------------------*)
  (* Implementation : Michael Riedl                                         *)
  (* Licence        : GNU Lesser General Public License (LGPL)              *)
  (*------------------------------------------------------------------------*)

  (* $Id: NumConvert.mod,v 1.2 2015/11/10 11:03:19 mriedl Exp mriedl $ *)

FROM SYSTEM     IMPORT TSIZE;
FROM StringsLib IMPORT Length,Copy;
FROM ConvTypes  IMPORT ConvResults;

(*========================== Lokale Objekte ================================*)

      (* 29.11.14 to replace constructs such as LONGCARD(MAX(INTEGER)) *)
      (*          +/-1 to avoid XDS code elimination message           *)

CONST MaxInt  = MAX(INTEGER)  - 1;
      MinInt  = MIN(INTEGER)  + 1;
      MaxCard = MAX(CARDINAL) - 1;

(* Wg. GM2  
 *    ConvStr = '0123456789ABCDEF';
 *)

VAR   ConvInt : ARRAY ['0'..'F'] OF SHORTCARD;
      ConvStr : ARRAY [0..15] OF CHAR;

CONST MaxBits = 8*TSIZE(CARDINAL);


PROCEDURE Check(VAR base : CARDINAL);

          (* Testet und korrigiert base *)
BEGIN
      IF (base < 2) THEN base := 2; ELSIF (base > 16) THEN base := 16; END;
END Check;

PROCEDURE Reverse(VAR Str      : ARRAY OF CHAR;
                      low,high : CARDINAL);

          (*--------------------------------------------------------*)
          (* Gibt die in Str enthaltene Zeichenkette in umgekehrter *)
          (* Reihenfolge zur"uck.                                   *)
          (*--------------------------------------------------------*)

          VAR chr : CHAR;
BEGIN
      WHILE (low < high) DO
        chr := Str[low]; Str[low] := Str[high]; Str[high] := chr;
        INC(low); DEC(high);
      END;
END Reverse;

PROCEDURE StrToLong(    Str    : ARRAY OF CHAR;
                        base   : CARDINAL;
                    VAR result : ConvResults) : LONGCARD;

          (*---------------------------------------------------------------*)
          (* Konvertiert den String Str in die entsprechende LONGCARD-Zahl *)
          (* Dabei wird ein f"uhrendes Vorzeichen, egal ob prositiv oder   *)
          (* negativ, ignoriert.                                           *)
          (* Unterroutine zu StrToIntBased und StrToCardBased.             *)
          (*---------------------------------------------------------------*)

          VAR  i,l : CARDINAL;
               b,t : LONGCARD;
               chr : CHAR;
               c   : SHORTCARD;
BEGIN
      Check(base);
      b := VAL( LONGCARD,base);
      i := 0;
      l := HIGH(Str);
      IF (Str[0] = '-') OR (Str[0] = '+') THEN i:=1; END;
      t := 0;
      IF (Str[i] = 0C) THEN result:=strEmpty; RETURN MAX(LONGCARD); END;
      WHILE (i <= l) AND (Str[i] # 0C) DO
        chr := Str[i];
        IF (chr < '0') OR (chr > 'F') THEN
          result:=strWrongFormat;
          RETURN MAX(LONGCARD);
        END;
        c := ConvInt[chr];
        IF (c > VAL(SHORTCARD,b) - 1) OR
        (t > (MAX(LONGCARD) - VAL(LONGCARD,c)) DIV b) THEN
          result:=strOutOfRange; (* Stimmt da\3 ?? *)
          RETURN MAX(LONGCARD);
        END;
        t := t*b + VAL(LONGCARD,c);
        INC(i);
      END;
      result:=strAllRight; 
      RETURN t;
END StrToLong;

(*====================== Globale Prozeduren. ===============================*)

PROCEDURE IntToStrBased(    I    : LONGINT;
                        VAR Str  : ARRAY OF CHAR;
                            base : CARDINAL;
                        VAR done : BOOLEAN);

          VAR  i,l : CARDINAL;
               b   : LONGINT; (* War LONGCARD, 28.11.14 *)
BEGIN
      done:=TRUE;
      l := HIGH(Str);
      Check(base);
      b := VAL(LONGCARD,base);
      IF (I < 0) THEN
        Str[0] := '-';
        i := 1; I := -I;
      ELSE
        i := 0;
      END;

      LOOP
        IF (i > l) THEN done:=FALSE; EXIT; END;
        Str[i] := ConvStr[I MOD b];
        INC(i);
        I := I DIV b;
        IF (I = 0) THEN EXIT END; (* ?? M.R. *)
      END;
      IF (i <= l) THEN Str[i]:=0C; END;
      IF (Str[0] < '0') THEN
        Reverse(Str,1,i-1);
      ELSE
        Reverse(Str,0,i-1);
      END;
END IntToStrBased;

PROCEDURE CardToStrBased(    C    : LONGCARD;
                         VAR Str  : ARRAY OF CHAR;
                             base : CARDINAL;
                         VAR done : BOOLEAN);

          VAR  i,l : CARDINAL;
               b   : LONGCARD;
BEGIN
      done:=TRUE;
      l := HIGH(Str);
      Check(base);
      b := VAL(LONGCARD,base);
      i := 0;
      LOOP
        IF (i > l) THEN done:=FALSE; EXIT END;
        Str[i] := ConvStr[( C MOD b )];
        INC(i);
        C := C DIV b;
        IF (C = 0) THEN EXIT END;
      END;
      IF (i <= l) THEN Str[i] := 0C; END;
      Reverse( Str,0,i-1 );
END CardToStrBased;

PROCEDURE StrToIntBased(    Str    : ARRAY OF CHAR;
                            base   : CARDINAL;
                        VAR I      : LONGINT;
                        VAR result : ConvResults);

          VAR C : LONGCARD;
BEGIN
      C := StrToLong(Str,base,result);
      IF (result # strAllRight) THEN
        I:=MAX(LONGINT);
      ELSIF (C > VAL(LONGCARD,MAX(LONGINT))) THEN
        result:=strOutOfRange; ;
        I:=MAX(LONGINT);
      ELSE
        I:=VAL(LONGINT,C);
        IF (Str[0] = '-') THEN I:= -I END;
      END;
END StrToIntBased;

PROCEDURE StrToCardBased(    Str    : ARRAY OF CHAR;
                             base   : CARDINAL;
                         VAR C      : LONGCARD;
                         VAR result : ConvResults);
BEGIN
      result:=strAllRight;
      C := StrToLong(Str,base,result);
      IF (Str[0] = '-') THEN result:=strWrongFormat; END;
END StrToCardBased;

(*===========================================================================*)

PROCEDURE IntToString(i      : INTEGER;
                      Breite : CARDINAL;
                  VAR Zahl   : ARRAY OF CHAR;
                  VAR done   : BOOLEAN);
BEGIN
      LongIntToString(VAL(LONGINT,i),Breite,Zahl,done);
END IntToString;

PROCEDURE StringToInt(    Zahl   : ARRAY OF CHAR;
                      VAR i      : INTEGER;
                      VAR result : ConvResults);

          VAR I : LONGINT;
BEGIN
      StrToIntBased(Zahl,10,I,result);
      IF (result # strAllRight) THEN
        i:=MAX(INTEGER);
      ELSIF (I > MaxInt) OR (I < MinInt) THEN
        result:=strOutOfRange;
        i:=MAX(INTEGER);
      ELSE
        i:=VAL(INTEGER,I);
      END;
END StringToInt;

PROCEDURE CardToString(    c      : CARDINAL;
                           Breite : CARDINAL;
                       VAR Zahl   : ARRAY OF CHAR;
                       VAR done   : BOOLEAN);
BEGIN
      LongCardToString(VAL(LONGCARD,c),Breite,Zahl,done);
END CardToString;

PROCEDURE StringToCard(    Zahl : ARRAY OF CHAR;
                       VAR c    : CARDINAL;
                       VAR result : ConvResults);

          VAR   C    : LONGCARD;
BEGIN
      StringToLongCard(Zahl,C,result);
      IF (result # strAllRight) THEN
        c:=MAX(CARDINAL);
      ELSIF (C > MaxCard) THEN
        result:=strOutOfRange;
        c:=MAX(CARDINAL);
      ELSE
        c:=VAL(CARDINAL,C);
      END;
END StringToCard;

PROCEDURE LongCardToString(    C      : LONGCARD;
                               Breite : CARDINAL;
                           VAR Zahl   : ARRAY OF CHAR;
                           VAR done   : BOOLEAN);

          VAR   ZahlStr        : ARRAY [0..31] OF CHAR;
                Len,diff,i,l,h : CARDINAL;
                Interim        : CHAR;
BEGIN
      done:=TRUE;
      IF (Breite > HIGH(Zahl)) THEN Breite:=HIGH(Zahl); END;
      Len:=0;
      REPEAT
        IF (Len = HIGH(Zahl)) THEN (* Fehler *)
          done:=FALSE;
          FOR i:=0 TO HIGH(Zahl)-1 DO Zahl[i]:="*"; END; Zahl[HIGH(Zahl)]:=0C;
          RETURN; 
        END;
        ZahlStr[Len]:=ConvStr[(C MOD 10)];
        INC(Len);
        C:=(C DIV 10);
      UNTIL (C = 0);
      ZahlStr[Len]:=0C;
      l:=0; h:=Len-1;
      WHILE (l < h) DO (* Umkehr der Zeichen *)
        Interim:=ZahlStr[l]; ZahlStr[l]:=ZahlStr[h]; ZahlStr[h]:=Interim;
        INC(l); DEC(h);
      END;
      IF (Len < Breite) THEN
        diff:=Breite - Len;
        FOR i:=0 TO diff-1 DO Zahl[i]:=' '; END;
        FOR i:=diff TO Breite DO Zahl[i]:=ZahlStr[i-diff]; END;
      ELSE
        Copy(Zahl,ZahlStr);
      END;
END LongCardToString;

PROCEDURE StringToLongCard(VAR ZahlStr : ARRAY OF CHAR;
                           VAR C       : LONGCARD;
                           VAR result  : ConvResults);
BEGIN
      StrToCardBased(ZahlStr,10,C,result);
END StringToLongCard;

PROCEDURE LongIntToString(    I      : LONGINT;
                              Breite : CARDINAL;
                          VAR Zahl   : ARRAY OF CHAR;
                          VAR done   : BOOLEAN);

          VAR ZahlStr    : ARRAY [0..31] OF CHAR;
              Len,diff,i : CARDINAL;
BEGIN
      IntToStrBased(I,ZahlStr,10,done);
      IF NOT done THEN RETURN; END;

      Len:=Length(ZahlStr);

      IF (Len < Breite) THEN
        diff:=Breite - Len;
        FOR i:=0 TO diff-1 DO Zahl[i]:=' '; END;
        FOR i:=diff TO Breite-1 DO Zahl[i]:=ZahlStr[i-diff]; END;
        Zahl[Breite]:=0C;
      ELSE
        Copy(Zahl,ZahlStr);
      END;
END LongIntToString;

PROCEDURE StringToLongInt(VAR ZahlStr : ARRAY OF CHAR;
                          VAR I       : LONGINT;
                          VAR result  : ConvResults);
BEGIN
      StrToIntBased(ZahlStr,10,I,result);
END StringToLongInt;

PROCEDURE CardToBitStr(    C      : CARDINAL;
                       VAR BitStr : ARRAY OF CHAR;
                           nbits  : CARDINAL);

          VAR   N,i,pot2  : CARDINAL;
BEGIN
      IF (nbits > MaxBits) THEN nbits:=MaxBits; END;
      IF (HIGH(BitStr) < (nbits-1)) THEN nbits:=HIGH(BitStr)-1; END;
      N:=nbits-1;
      (* take care that 2**i in FOR loop does not exceed range *)
      IF (nbits = MaxBits) THEN DEC(N); END;
      FOR i:=0 TO nbits-1 DO BitStr[i]:="0"; END;
      BitStr[nbits]:=0C;
      IF ODD (C) THEN BitStr[0]:="1"; END;
      pot2:=2;
      FOR i:=1 TO N DO
        IF ODD (C DIV pot2) THEN BitStr[i]:="1"; END;
        pot2:=2*pot2;
        IF (pot2 > C) THEN RETURN; END; (* All remaining bits are zero *)
      END;
      (* nbits = MaxBits, check high bit if not already exited  *)
      IF ODD (C DIV pot2) THEN BitStr[MaxBits-1]:="1"; END;
END CardToBitStr;

VAR   c : CHAR;

BEGIN
      FOR c:="0" TO "F" DO ConvInt[c] := MAX(SHORTCARD); END;

      ConvInt["0"] :=  0; ConvInt["1"] :=  1; ConvInt["2"] :=  2;
      ConvInt["3"] :=  3; ConvInt["4"] :=  4; ConvInt["5"] :=  5;
      ConvInt["6"] :=  6; ConvInt["7"] :=  7; ConvInt["8"] :=  8;
      ConvInt["9"] :=  9;
      ConvInt["A"] := 10; ConvInt["B"] := 11; ConvInt["C"] := 12;
      ConvInt["D"] := 13; ConvInt["E"] := 14; ConvInt["F"] := 15;

      ConvStr := '0123456789ABCDEF';
END NumConvert.
