IMPLEMENTATION MODULE RealConvert;

  (*------------------------------------------------------------------------*)
  (* Modul fuer die Konvertierung Fliesskommazahlen                         *)
  (* Modul for the conversion of real numbers to string or visa versa.      *)
  (*------------------------------------------------------------------------*)
  (* Letzt Bearbeitung:                                                     *)
  (*                                                                        *)
  (* 03.10.95, MRi: Korrekturen                                             *)
  (* 22.02.15, MRi: Externes TestReal                                       *)
  (* 14.06.15, MRi: Ergaenzung in StringToReal so daß Fortran D-Exponent    *)
  (*                akzeptiert wird (war in IstZulaessig(chr) vorbereitet   *)
  (*                aber nicht in StringToReal                              *)
  (* 30.10.15, MRi: Einfuehren von SigFig in RealToFixString, Ziffern nun   *)
  (*                als Konstante (statt VAR - Parameter), MachEps          *)
  (*                eingefuegt.                                             *)
  (* 12.01.18, MRi: Ziffern wieder VAR-Parameter wegen GNU M2               *)
  (*------------------------------------------------------------------------*)
  (* Offene Punkte:                                                         *)
  (*                                                                        *)
  (* - In RealToFloatString Abfragen des Parameters ExpLen ueberpruefen     *)
  (*------------------------------------------------------------------------*)
  (* Implementation : Michael Riedl                                         *)
  (* Licence        : GNU Lesser General Public License (LGPL)              *)
  (*------------------------------------------------------------------------*)

  (* $Id: RealConvert.mod,v 1.2 2015/10/30 14:48:14 mriedl Exp mriedl $ *)

FROM SYSTEM    IMPORT BYTE,ADR;
FROM Errors    IMPORT ErrOut;
               IMPORT Errors;
FROM ConvTypes IMPORT ConvResults;
               IMPORT StringsLib,NumConvert;
               IMPORT TestReal;

CONST     debug   = FALSE; (* Fehlerausgaben auf den Bildschirm , wenn ein *)
                           (* Fehler aufgetreten ist.                      *)

CONST  (* Ziffern = "0123456789"; wg. GM2 *)
          MachEps = 2.22044604925031308E-16; (* Ggf. anpassen *)
          SigFig  = 16;                      (* Abs. Exponent von MachEps *)

VAR       RKonv        : ARRAY ['0'..'9'] OF LONGREAL;
          IKonv        : ARRAY ['0'..'9'] OF CARDINAL;
          KommaZeichen : CHAR; (* Macht 1,0 statt 1.0 in Ausgaben m"oglich. *)
          Ziffern      : ARRAY [0..11] OF CHAR;

(*======================== Lokale Prozeduren =============================*)

PROCEDURE CardPow(x : LONGREAL; (* Indentisch mit LMathLib.CardPot. *)
                  i : CARDINAL) : LONGREAL;

          VAR     z : LONGREAL;
BEGIN
      z:=1.0;
      LOOP
        IF ODD(i) THEN z:=z*x END;
        i:=i DIV 2; IF (i = 0) THEN EXIT END;
        x:=x*x;
      END; (* LOOP *)
      RETURN z;
END CardPow;

PROCEDURE IstZulaessig(chr : CHAR) : BOOLEAN;

          (* Pr"uft, ob chr ein zulaessiges Zeichen innerhalb eines *)
          (* Realstrings ist.                                       *)  

BEGIN
      CASE chr OF
        ".",",",
        "E","e",
        "D","d",
        "0".."9" : RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
END IstZulaessig;

(*========================================================================*)

PROCEDURE StringToReal(VAR Zahl   : ARRAY OF CHAR;
                       VAR x      : LONGREAL;
                       VAR result : ConvResults);

         VAR i,j,ik,exp,BasLen  : INTEGER; (* Ver"andert von CARDINAL *)
             BasStr,ExpStr      : POINTER TO ARRAY [0..255] OF CHAR;
             Neg,Komma,Exponent : BOOLEAN;
             Chr                : CHAR;

BEGIN
      result:=strAllRight;
      i:=0; WHILE (Zahl[i] = ' ') DO INC(i);  END;
      Neg:=FALSE;
      IF (Zahl[i] = '-') THEN
        Neg:=TRUE; INC(i);
      ELSIF (Zahl[i] = '+') THEN
        INC(i);
      END;
      BasStr:=ADR(Zahl[i]);
      j:=0;
      WHILE (CAP(BasStr^[j]) # 'E') AND 
            (CAP(BasStr^[j]) # 'D') AND (* Fortran double precision *)
            (BasStr^[j]      #  0C) AND
            (BasStr^[j]      # " ")
      DO
        IF NOT IstZulaessig(BasStr^[j]) THEN
          IF debug THEN
            Errors.WriteString("FEHLER in RealConvert.StringToReal : ");
            Errors.WriteString("unzulaessiges Zeichen [");
            Errors.WriteChar(BasStr^[j]); 
            Errors.WriteChar(']'); 
            Errors.WriteLn; 
          END;
          result:=strWrongFormat;
          x := MAX(LONGREAL);
          RETURN;
        END;
        INC(j); 
      END;
      BasLen:=j;
      Exponent:=FALSE;
      IF (BasStr^[j] # 0C) AND (BasStr^[j] # " ") THEN
        ExpStr:=ADR(Zahl[i+j+1]);
        Exponent:=TRUE;
      END;

      x:=0.0; Komma:=FALSE; ik:=0; (* Auswerten der Basis *)
      FOR i:=0 TO BasLen-1 DO
        Chr:=BasStr^[i];
        IF (Chr = KommaZeichen) OR (Chr = '.') THEN (* Doppelte Abfrage? *)
          IF (i # BasLen-1) THEN Komma:=TRUE; END;
        ELSIF (Chr >= '0') AND (Chr <= '9') THEN
          x := x * 10.0 + RKonv[Chr];
          IF Komma THEN INC(ik); END;
        ELSE
          result:=strWrongFormat; x:=MAX(LONGREAL); RETURN;
        END;
      END;
      IF Komma THEN x := x / CardPow(10.0,ik); END;
      IF Neg THEN x:=-x; END;
      IF Exponent THEN (* Auswerten den Exponenten *)
        Neg:=FALSE; i:=0;
        IF (ExpStr^[0] = '-') THEN
          Neg:=TRUE; INC(i);
        ELSIF (ExpStr^[0] = '+') THEN
          INC(i);
        END;
        exp:=0;
        WHILE (ExpStr^[i] # 0C) DO
          IF (ExpStr^[i] < '0') OR (ExpStr^[i] > '9') THEN
            result:=strWrongFormat; x:=MAX(LONGREAL); RETURN;
          END;
          exp:=exp * 10 + VAL(INTEGER,IKonv[ExpStr^[i]]);
          INC(i);
        END;
        IF Neg THEN x:=x / CardPow(10.0,exp) ELSE x:=x * CardPow(10.0,exp) END;
      END;
END StringToReal;

PROCEDURE FehlStr(VAR Str : ARRAY OF CHAR;
                      len : CARDINAL);

          VAR i : CARDINAL;

BEGIN
      IF (len > HIGH(Str)) THEN len:=HIGH(Str); END;
      FOR i:=0 TO len-1 DO Str[i]:='*'; END; Str[len]:=0C;
END FehlStr;

PROCEDURE RealToFixString(VAR RealStr : ARRAY OF CHAR;
                              x       : LONGREAL;
                              Breite  : CARDINAL;
                              Komma   : CARDINAL;
                          VAR done    : BOOLEAN);

          CONST eps           = MachEps;

          VAR   Index,vor,sig : CARDINAL; (* sig = signifikante Ziffern *)
                i,j,k         : CARDINAL;
                leer,l        : INTEGER;
                ABSx          : LONGREAL;
BEGIN
      IF (Breite > HIGH(RealStr)) THEN Breite:=HIGH(RealStr); END;
      IF TestReal.IsReal8NaN(x) THEN
        StringsLib.Copy(RealStr,"  NAN");
        FOR i:=5 TO Breite-1 DO RealStr[i]:=" "; END; RealStr[Breite]:=0C;
        done:=FALSE;
        RETURN;
      END;
      IF TestReal.IsReal8INF(x) THEN
        IF (x < 0.0) THEN
          StringsLib.Copy(RealStr," -INF");
        ELSE
          StringsLib.Copy(RealStr,"  INF");
        END;
        FOR i:=5 TO Breite-1 DO RealStr[i]:=" "; END; RealStr[Breite]:=0C;
        done:=FALSE;
        RETURN;
      END;
      done:=TRUE;
      ABSx:=ABS(x); vor:=0; (* Z"ahle Vorkommastellen *)
      WHILE ((ABSx - 1.0) >= -10.0*eps) DO ABSx:=0.1*ABSx; INC(vor); END;

      IF (vor > Breite-2) THEN FehlStr(RealStr,Breite); done:=FALSE; RETURN END;
      IF (Komma > (Breite - vor - 2)) THEN Komma := Breite - vor - 2; END;
      (* Stimmt hier die Reihenfolge der Abfragen ? *)
      ABSx:=ABSx + 0.5*CardPow(0.1,(Komma+vor));   (* Runden *)
      IF (ABSx >= 1.0) THEN ABSx:=0.1*ABSx; INC(vor) END; (* Wichtig !! *)

      leer := Breite - Komma - vor - 2;
      IF (vor = 0) THEN DEC(leer); END;

      sig := 0; (* Zaehlt die Praeezision *)
      i:=0; (* i z"ahlt Position im RealStr *)
      IF (leer > 0) AND (leer < VAL(INTEGER,HIGH(RealStr))) THEN
        FOR l:=0 TO leer-1 DO RealStr[l]:=' '; END; (* F"uhrende Leerzeichen *)
        i:=VAL(CARDINAL,leer);
      END;
      IF (x >= 0.0) THEN RealStr[i]:=' '; ELSE RealStr[i]:='-'; END; INC(i);
      IF (vor = 0) THEN RealStr[i]:='0'; INC(i); END;
      WHILE (vor > 0) DO (* Ganzzahliger Anteil der Zahl *)
        IF (sig < SigFig) THEN ABSx:=10.0*ABSx; INC(sig); END;
        Index:=TRUNC(ABSx);
        IF (Index > 9) THEN
          Errors.ErrOut(
          "in RealConvert.RealToFixString (Vorkommaanteil) !");
          done:=FALSE; FehlStr(RealStr,Breite);
          RETURN;
        END;
        RealStr[i]:=Ziffern[Index];
        DEC(vor); INC(i);
        ABSx:=ABSx - LFLOAT(Index);
      END;
      IF (i > HIGH(RealStr)) THEN
        FehlStr(RealStr,Breite); done:=FALSE; RETURN
      END;
      RealStr[i]:=KommaZeichen; INC(i);
      IF (Komma > 0) AND (i < HIGH(RealStr)) THEN
        (* Nachkommastellen der Zahl *)
        FOR j:=1 TO Komma DO
          IF (sig < SigFig) THEN ABSx:=10.0*ABSx; INC(sig); END;
          Index:=TRUNC(ABSx);
          IF (i > HIGH(RealStr)) THEN
            Errors.ErrOut(
            "Zeichenkette zu kurz (RealConvert.RealToFixString) !");
            done:=FALSE;
            RealStr[Breite]:=0C;
            RETURN;
          END;
          IF (Index > 9) THEN
            IF debug THEN
              Errors.WriteString("Index : "); 
              Errors.WriteCard(Index);
              Errors.WriteLn; 
              Errors.ErrOut(
              "in RealConvert.RealToFixString (Nachkommaanteil) !");
            END;
            done:=FALSE;
            FOR k:=i TO Breite DO RealStr[k]:="*"; END; RealStr[Breite]:=0C;
            RETURN;
          END;
          RealStr[i]:=Ziffern[Index];
          INC(i);
          ABSx:=ABSx - LFLOAT(Index);
        END;
      END;
      RealStr[i]:=0C;
END RealToFixString;

PROCEDURE RealToFloatString(VAR RealStr : ARRAY OF CHAR;
                                x       : LONGREAL;
                                Breite  : CARDINAL; (* n : Feldbreite       *)
                                Komma   : INTEGER;  (* m : Nachkommastellen *)
                                ExpLen  : CARDINAL; (* 2,3 Exponentstellen  *)
                            VAR done    : BOOLEAN);

          (*---------------------------------------------------------------*)
          (* Stark "uberarbeitete Realausgaberoutine von :                 *)
          (* Maier,G.;    Hybridrechenzentrum, ETH Z"urich (PDP11 ?),      *)
          (* Lutz,J.;     Inst. f. Informationsv. Uni T"ubingen (Unix) und *)
          (* Kupfer,J.P.; (SIGMA) (f"ur Amiga)                             *)
          (*                                                               *)
          (* Die Routine schreibt nicht mehr in eine Datei, sondern gibt   *)
          (* ihr Ergebniss in einem String zur"uck. Zudem kann man nun die *)
          (* Exponentenfeldbreite ausw"ahlen.                              *)
          (*---------------------------------------------------------------*)

  PROCEDURE max(a,b:INTEGER) : INTEGER;

  BEGIN
        IF a >= b THEN RETURN a ELSE RETURN b END;
  END max;

  PROCEDURE exp10(i : INTEGER) : LONGREAL;

  BEGIN
        IF (i >= 0) THEN
          RETURN CardPow(10.0,VAL(CARDINAL,i));
        ELSE
          RETURN 1.0 / CardPow(10.0,VAL(CARDINAL,ABS(i)));
        END;
  END exp10;

  VAR     IPos       : CARDINAL; (* Zaehlt Position in RealStr (M.R.) *)
          LocRealStr : ARRAY [0..64] OF CHAR; (* M.R. *)

  PROCEDURE convertnumber(num,len : CARDINAL;
                          ch      : CHAR;
                          neg     : BOOLEAN);

            VAR    digit : ARRAY[1..9] OF CHAR;
                   i,cnt : CARDINAL;

  BEGIN
        cnt:=0;
        REPEAT
          INC(cnt);
(*        digit[cnt] := CHAR((num MOD 10) + CARDINAL('0')); *)
          digit[cnt] := Ziffern[num MOD 10]; (* M.R. *)
          num := num DIV 10;
        UNTIL num=0;
        IF neg THEN INC(cnt); digit[cnt]:='-' END;
        FOR i:=len TO cnt+1 BY -1 DO LocRealStr[IPos] := ch; INC(IPos); END;
        FOR i:=cnt TO 1 BY -1 DO
          LocRealStr[IPos] := digit[i]; INC(IPos);
        END;
  END convertnumber;


      VAR  exponent                   : INTEGER;
           before,behind,minimal,i,ix : CARDINAL;
           deltaround                 : LONGREAL;
           sign,exponential           : BOOLEAN;
           ch                         : CHAR;

BEGIN
      IF TestReal.IsReal8NaN(x) THEN
        StringsLib.Copy(RealStr,"  NAN");
        FOR i:=5 TO Breite-1 DO RealStr[i]:=" "; END; RealStr[Breite]:=0C;
        done:=FALSE;
        RETURN;
      END;
      IF TestReal.IsReal8INF(x) THEN
        IF (x < 0.0) THEN
          StringsLib.Copy(RealStr," -INF");
        ELSE
          StringsLib.Copy(RealStr,"  INF");
        END;
        FOR i:=5 TO Breite-1 DO RealStr[i]:=" "; END; RealStr[Breite]:=0C;
        done:=FALSE;
        RETURN;
      END;
      (* Testen *)
      IF (ExpLen > 4) THEN ExpLen:=4; ELSIF (ExpLen < 1) THEN ExpLen:=1; END;
(*
      IF (ExpLen > 3) THEN ExpLen:=3; ELSIF (ExpLen < 2) THEN ExpLen:=2; END;
 *)
      (* Vorzeichen , x dann vorzeichenlos ! *)
      sign := (x < 0.0); x:=ABS(x);
      IF (Komma < 0) OR (x >= 1.0E+08) THEN (* Exponentialdarstellung ? *)
        exponential := TRUE;
        Komma       := ABS(Komma);
        IF (ExpLen <= 2) AND ((x >= 1.0E+100) OR (x <= 1.0E-100)) AND (x # 0.0)
        THEN
          ExpLen:=3; DEC(Komma);
(*        ErrOut('ExpLen ge"andert (RealConvert.RealToFloatString)!'); *)
        END;
        minimal     := 5 + ExpLen; (* Minimale Zeichenanzahl *)
      ELSE
        exponential := FALSE;
        minimal     := 3; (* Zeichen *)
      END; (* IF *)

      (* Ermittle, wieviele Zeichen vor und nach dem Dezimalkomma *)
      (* ben"otigt werden.                                        *) 

      Breite := max(Breite,minimal);
      behind := -max(-Komma,VAL(INTEGER,minimal) - VAL(INTEGER,Breite));
      before := Breite - behind - minimal + 2;
      deltaround := 0.5*exp10(-VAL(INTEGER,behind));
      IF exponential THEN (* normalize *)
        exponent := 0;
        WHILE (x < (1.0 - deltaround)) AND (x # 0.0) DO
          x := x*10.0; DEC(exponent);
        END;
        WHILE (x >= (10.0 - deltaround)) DO
          x := x*0.1; INC(exponent);
        END;
      END; (* IF *)
      x := x + deltaround; (* Runden *)

      IPos:=0; (* Initialisierung, Position im String *)

      ch := ' ';  (* Stellen vor dem Dezimalpunkt *)
      IF (x >= 1.0E+04) THEN   (* Wieso gerade 4 ??? *)
        ix := TRUNC(x*1.0E-04);
        convertnumber(ix,max(before-4,1),ch,sign);
        sign := FALSE;
        before := 4;
        x:=x - LFLOAT(ix)*1.0E+04;
        ch := '0';
      END; (* IF *)
      ix := TRUNC(x);
      convertnumber(ix,before,ch,sign);
      x := x - LFLOAT(ix);

      LocRealStr[IPos]:= KommaZeichen; INC(IPos);

      WHILE (behind > 0) DO (* Stellen nach dem Dezimalpunkt *)
        i := -max(-VAL(INTEGER,behind),-4);
        x := x*exp10(i); ix := TRUNC(x);
        convertnumber(ix,i,'0',FALSE);
        x := x - LFLOAT(ix);
        DEC(behind,i);
      END; (* WHILE *)

      IF exponential THEN (* Exponent *)
        LocRealStr[IPos]:= "E"; INC(IPos);
        IF (exponent >= 0) THEN
          LocRealStr[IPos]:= "+"; INC(IPos);
        ELSE
          LocRealStr[IPos]:= "-"; INC(IPos);
        END;
        convertnumber(ABS(exponent),ExpLen,'0',FALSE);
      END; (* IF *)

      i:=0;                  (* Umkopieren auf "Ubergabeparameter, test ob *)
      WHILE (i < IPos) DO    (* L"ange in Ordnung (M.R.) *)
        RealStr[i]:=LocRealStr[i];
        INC(i);
        IF (i >= HIGH(RealStr)) THEN
          FehlStr(RealStr,Breite);
          done:=FALSE; RETURN;
        END;
      END;
      RealStr[i]:=0C;
      done:=TRUE;
END RealToFloatString;

PROCEDURE RealToEngString(VAR RStr   : ARRAY OF CHAR;
                              x      : LONGREAL;
                              Breite : CARDINAL;
                              Komma  : CARDINAL;
                              ExpLen : CARDINAL;
                          VAR done   : BOOLEAN);

          VAR Exp    : INTEGER;
              i      : CARDINAL;
              ExpStr : ARRAY [0..15] OF CHAR;

BEGIN
      Exp:=0;
      IF (ABS(x) >= 1000.0) THEN
        REPEAT
          INC(Exp); x:=0.1*x;
        UNTIL ((Exp MOD 3) = 0) AND (ABS(x) < 1000.0);
      ELSIF (ABS(x) < 1.0) AND (x # 0.0) THEN
        REPEAT
          DEC(Exp); x:=10.0*x;
        UNTIL ((ABS(Exp) MOD 3) = 0) AND (Exp # 0) AND (ABS(x) >= 1.0);
      END;
      IF (Exp > 0) THEN
        x:=x * CardPow(1.0,VAL(CARDINAL,Exp));
      ELSE
        x:=x / CardPow(1.0,VAL(CARDINAL,-Exp));
      END;

      IF (Breite < ExpLen + 3) THEN done:=FALSE; FehlStr(RStr,Breite); END;
      RealToFixString(RStr,x,Breite-ExpLen-2,Komma,done);
      IF NOT done THEN FehlStr(RStr,Breite); RETURN; END;
      NumConvert.IntToString(ABS(Exp),ExpLen,ExpStr,done);
      IF NOT done THEN FehlStr(RStr,Breite); RETURN; END;
      FOR i:=0 TO ExpLen-1 DO
        IF (ExpStr[i] = " ") THEN ExpStr[i]:="0"; END;
      END;
      IF (Exp < 0) THEN StringsLib.Append(RStr,"E-"); ELSE StringsLib.Append(RStr,"E+"); END;
      StringsLib.Append(RStr,ExpStr);
END RealToEngString;

VAR       c          : CHAR;
          i          : CARDINAL;

BEGIN
      i:=0;
      FOR c:='0' TO '9' DO RKonv[c]:=LFLOAT(i); IKonv[c]:=i; INC(i); END;
      KommaZeichen:='.'; 
      Ziffern := "0123456789"; (* wg. GM2 *)
END RealConvert.
