IMPLEMENTATION MODULE TIO; 

  (*------------------------------------------------------------------------*)
  (* Bildschirmorientierte Lese- und Schreibprozeduren.                     *)
  (* Terminal oriented read and write procedures.                           *)
  (*------------------------------------------------------------------------*)
  (*                                                                        *)
  (* Letzte Aenderung                                                       *)
  (*                                                                        *)
  (* 30.03.97, MRi: Korrekturen                                             *)
  (* 19.02.15, MRi: Prozedure WrBits eingefuegt                             *)
  (* 04.05.15, MRi: Prozedure WrLngInt und WrReal eingefuegt                *)
  (* 04.02.16, MRi: Prozedure RdLngInt eingefuegt                           *)
  (* 26.07.17. MRi: Verweise auf Record-Complex-Typen geloescht             *)
  (* 26.07.17, MRi: Verweise auf Record-Complex-Typen geloescht             *)
  (* 03.10.17, MRi: In WrLngCmply Korrektur an ExpAnz vorgenommen           *)
  (*------------------------------------------------------------------------*)
  (* Offene Punkte                                                          *)
  (*                                                                        *)
  (* - Die Varibale ExpAnz in einer Prozedur nach aussen kapseln            *)
  (*------------------------------------------------------------------------*)
  (* Implementation : Michael Riedl                                         *)
  (* Licence        : GNU Lesser General Public License (LGPL)              *)
  (*------------------------------------------------------------------------*)

  (* $Id: TIO.mod,v 1.1 2015/09/13 20:48:43 mriedl Exp mriedl $ *)

                 IMPORT ASCII,CommonIO;
FROM FileSystem  IMPORT Input,Output,ReadNBytes;
FROM Errors      IMPORT Fehler,Fehlerflag,ErrOut;
FROM ConvTypes   IMPORT ConvResults;
FROM StringsLib  IMPORT Concat,Length,Copy;
FROM NumConvert  IMPORT StringToInt,StringToLongInt,
                        StringToCard,StringToLongCard,
                        CardToStrBased,StrToCardBased,CardToBitStr;
FROM RealConvert IMPORT StringToReal;
                 IMPORT FIO2;

(*======================= Module-globale Objekte ==========================*)

CONST     MaxCardStr    = 15;
          MaxLngCardStr = 31;
          MaxIntStr     = 15;
          MaxLngIntStr  = 31;
          MaxRealStr    = 31;
          Return        = CHR(10); (* ASCII.lf ? *)

VAR       RealStr       : ARRAY [0..MaxRealStr] OF CHAR;
          CardStr       : ARRAY [0..MaxCardStr] OF CHAR;
          LngCardStr    : ARRAY [0..MaxLngCardStr] OF CHAR;
          LngIntStr     : ARRAY [0..MaxLngIntStr] OF CHAR;
          IntStr        : ARRAY [0..MaxIntStr]  OF CHAR;
          result        : ConvResults;
          FFlag         : ARRAY [strAllRight..strEmpty] OF
                          ARRAY [0..31] OF CHAR;

PROCEDURE RdWort(VAR Wort : ARRAY OF CHAR);

          (*----------------------------------------------------*)
          (* Lie\3t ein einzelnes Wort aus der Standarteingabe. *)
          (*----------------------------------------------------*)

          VAR i   : CARDINAL;
              Chr : CHAR;

BEGIN
      REPEAT
        RdChar(Chr);
        IF (Chr = Return) THEN Wort[0]:=0C; RETURN; END;
      UNTIL NOT (Chr = ' ');
      Wort[0]:=Chr; i:=1;
      LOOP
        IF (i = HIGH(Wort)) THEN EXIT END;
        RdChar(Chr);
        IF (Chr = Return) THEN EXIT END;
        Wort[i]:=Chr; INC(i);
      END;
      Wort[i]:=0C;
END RdWort;

(*======================= Leseprozeduren ==================================*)

PROCEDURE RdChar(VAR chr : CHAR);

          VAR i : CARDINAL;
BEGIN
      ReadNBytes(Input,chr,1,i);
END RdChar;

PROCEDURE RdStr(VAR Wort : ARRAY OF CHAR);

          VAR  i   : CARDINAL;
               chr : CHAR;
BEGIN
      i:=0;
      LOOP
        RdChar(chr);
        IF (chr = Return) THEN EXIT END;
        Wort[i]:=chr;
        INC(i);
        IF (i = HIGH(Wort)) THEN EXIT END;
      END; (* LOOP *)
      Wort[i]:=0C;
END RdStr;

PROCEDURE RdCard(VAR c : CARDINAL);

BEGIN
      RdWort(CardStr);
      StringToCard(CardStr,c,result);
      IF (result # strAllRight) THEN
        Concat(Fehlerflag,FFlag[result]," (TIO.RdCard) !");
        ErrOut(Fehlerflag);
        Fehler:=TRUE;
        c:=MAX(CARDINAL);
        RETURN;
      END;
END RdCard;

PROCEDURE RdHex(VAR c : CARDINAL);

          VAR  CardStr : ARRAY [0..15] OF CHAR;
               C       : LONGCARD;

BEGIN
      RdWort(CardStr);
      StrToCardBased(CardStr,16,C,result);
      IF (result # strAllRight) THEN
        Concat(Fehlerflag,FFlag[result]," (Streams.RdCard) !");
        ErrOut(Fehlerflag);
        Fehler:=TRUE;
        c:=MAX(CARDINAL);
        RETURN;
      END;
      IF (C > VAL(LONGCARD,MAX(CARDINAL))) THEN
        Concat(Fehlerflag,FFlag[strOutOfRange]," (Streams.RdHex) !");
        ErrOut(Fehlerflag);
        c:=MAX(CARDINAL);
        RETURN;
      END;
      c := VAL(CARDINAL,C);
END RdHex;

PROCEDURE WrHex(c : CARDINAL;
                n : CARDINAL);  (* Feldbreite *)

          VAR ZahlStr : ARRAY [0..15] OF CHAR;
              done    : BOOLEAN;
              len,i   : CARDINAL;

BEGIN
       CardToStrBased(VAL(LONGCARD,c),ZahlStr,16,done);
       len := Length(ZahlStr);
       IF (n > len) THEN
         FOR i:=1 TO (n - len) DO WrChar(" "); END;
       END;
       WrStr(ZahlStr);
END WrHex;

PROCEDURE WrBits(c     : CARDINAL;
                 nbits : CARDINAL);  (* Feldbreite *)
          
          VAR Str : ARRAY [0..127] OF CHAR;
BEGIN
      CardToBitStr(c,Str,nbits); 
      WrStr(Str);
END WrBits;

PROCEDURE RdLngCard(VAR C : LONGCARD);

BEGIN
      RdWort(LngCardStr);
      StringToLongCard(LngCardStr,C,result);
      IF (result # strAllRight) THEN
        Concat(Fehlerflag,FFlag[result]," (TIO.RdLngCard) !");
        ErrOut(Fehlerflag);
        Fehler:=TRUE;
        C:=MAX(LONGCARD);
        RETURN;
      END;
END RdLngCard;

PROCEDURE RdInt(VAR i : INTEGER);

BEGIN
      RdWort(IntStr);
      StringToInt(IntStr,i,result);
      IF (result # strAllRight) THEN
        Concat(Fehlerflag,FFlag[result]," (TIO.RdInt) !");
        ErrOut(Fehlerflag);
        Fehler:=TRUE;
        i:=MAX(INTEGER);
        RETURN;
      END;
END RdInt;

PROCEDURE RdLngInt(VAR I : LONGINT);

BEGIN
      RdWort(LngIntStr);
      StringToLongInt(LngCardStr,I,result);
      IF (result # strAllRight) THEN
        Concat(Fehlerflag,FFlag[result]," (TIO.RdLngCard) !");
        ErrOut(Fehlerflag);
        Fehler:=TRUE;
        I:=MAX(LONGINT);
        RETURN;
      END;
END RdLngInt;

PROCEDURE RdLngReal(VAR x : LONGREAL);

BEGIN
      RdWort(RealStr);
      StringToReal(RealStr,x,result);
      IF (result # strAllRight) THEN
        Concat(Fehlerflag,FFlag[result]," (TIO.RdLngReal) !");
        ErrOut(Fehlerflag);
        Fehler:=TRUE;
        x:=MAX(LONGREAL);
      END;
END RdLngReal;

(*======================= Schreibprozeduren ===============================*)

PROCEDURE WrLn();

BEGIN
      IF CommonIO.CrLf THEN FIO2.WrChar(Output,ASCII.cr); END;
      FIO2.WrChar(Output,ASCII.lf);
END WrLn;

PROCEDURE WrLnN(n : CARDINAL);  (* Anzahl der Zeilenumbr\"uche. *)

          VAR         i     : CARDINAL;
BEGIN
      FOR i:=1 TO n DO FIO2.WrLn(Output); END;
END WrLnN;

PROCEDURE WrChar(chr : CHAR);

BEGIN
      IF (chr = CommonIO.EoL) THEN
        IF CommonIO.CrLf THEN FIO2.WrChar(Output,ASCII.cr); END;
        FIO2.WrChar(Output,ASCII.lf);
      ELSE
        FIO2.WrChar(Output,chr);
      END;
END WrChar;

PROCEDURE WrCharN(chr    : CHAR;
                  Anzahl : CARDINAL);

          CONST puflen = 31;

          VAR puf : ARRAY [0..puflen] OF CHAR; (* Puffer *)
              i,j : CARDINAL;

BEGIN
      WHILE (Anzahl > 0) DO
        i :=puflen;
        IF (i > Anzahl) THEN i := Anzahl; END;
        DEC(Anzahl,i);
        j := 0;
        WHILE (j < i) DO puf[j] := chr; INC(j) END;
        puf[j] := 0C;
        FIO2.WrStr(Output,puf);
      END;
END WrCharN;

PROCEDURE WrStr(Wort : ARRAY OF CHAR);

BEGIN
      FIO2.WrStr(Output,Wort);
END WrStr;

PROCEDURE WrStrForm(Str    : ARRAY OF CHAR;
                    Breite : INTEGER);

          VAR  len   : CARDINAL;
               fuell : INTEGER;

BEGIN
      len := Length(Str);
      fuell := ABS(Breite) - VAL(INTEGER,len);
      IF (Breite > 0) AND (fuell > 0) THEN WrCharN(' ',fuell); END;
      FIO2.WrStr(Output,Str);
      IF (Breite < 0) AND (fuell > 0) THEN WrCharN(' ',fuell); END;
END WrStrForm;

PROCEDURE WrCard(c : CARDINAL;
                 n : CARDINAL);  (* Feldbreite *)
BEGIN
      FIO2.WrCard(Output,c,n);
END WrCard;

PROCEDURE WrLngCard(C : LONGCARD;
                    n : CARDINAL);  (* Feldbreite *)
BEGIN
       FIO2.WrLngCard(Output,C,n);
END WrLngCard;

PROCEDURE WrInt(i : INTEGER;
                n : CARDINAL);  (* Feldbreite *)
BEGIN
      FIO2.WrInt(Output,i,n);
END WrInt;

PROCEDURE WrLngInt(I : LONGINT;
                   n : CARDINAL);  (* Feldbreite *)
BEGIN
      FIO2.WrLngInt(Output,I,n);
END WrLngInt;

PROCEDURE WrLngReal(x      : LONGREAL;
                    Breite : CARDINAL; (* n : Feldbreite       *)
                    Komma  : INTEGER); (* m : Nachkommastellen *)

          VAR interim : CARDINAL;
BEGIN
      interim:=FIO2.ExpAnz; FIO2.ExpAnz:=ExpAnz;
      FIO2.WrLngReal(Output,x,Breite,Komma);
      FIO2.ExpAnz:=interim;
END WrLngReal;

PROCEDURE WrReal(x      : REAL;
                 Breite : CARDINAL; (* Feldbreite       *)
                 Komma  : INTEGER); (* Nachkommastellen *)

          VAR interim : CARDINAL;
BEGIN
      interim:=FIO2.ExpAnz; FIO2.ExpAnz:=2;
      FIO2.WrLngReal(Output,VAL(LONGREAL,x),Breite,Komma);
      FIO2.ExpAnz:=interim;
END WrReal;

PROCEDURE WrLngCmplx(x      : LONGCOMPLEX;
                     Breite : CARDINAL; (* Feldbreite       *)
                     Komma  : INTEGER); (* Nachkommastellen *)

          VAR interim : CARDINAL;
BEGIN
      interim:=FIO2.ExpAnz; FIO2.ExpAnz:=ExpAnz;
      FIO2.WrLngReal(Output,RE(x),Breite,Komma);
      IF (IM(x) >= 0.0) THEN
        FIO2.WrStr(Output,' + i');
      ELSE
        FIO2.WrStr(Output,' - i');
      END;
      FIO2.WrLngReal(Output,ABS(IM(x)),Breite,Komma);
      FIO2.ExpAnz:=interim;
END WrLngCmplx;

PROCEDURE WrBool(Bool  : BOOLEAN);
BEGIN
      IF Bool THEN WrStr(' TRUE  '); ELSE WrStr(' FALSE '); END;
END WrBool;

PROCEDURE RdRealDefault(VAR X : LONGREAL;   (* R"uckgabewert *)
                            Y : LONGREAL);  (* Default-Wert *)

          (*-------------------------------------------------*)
          (* Lie3t eine Realzahl vom Terminal.               *)
          (* Falls sofort RETURN gegeben wird, wird Y als    *)
          (* Default-Wert auf X zur"uckgegeben. Bei Eingabe- *)
          (* fehlern wird solange widerholt, bis eine legale *)
          (* Eingabe erfolgt.                                *)
          (*-------------------------------------------------*)

PROCEDURE Test(ch : CHAR) : BOOLEAN;

          (*------------------------------------------------------------------*)
          (* Testet ob ch ein f"ur Real-Zahlen zul"assiges Stringelement ist. *)
          (*------------------------------------------------------------------*)

BEGIN
          CASE ch OF 
            "0".."9","+","-",".","e","E" : RETURN TRUE;
          ELSE  (* ch # '0..9' 'e' 'E' '+' '-' '.' *)
            RETURN FALSE;
          END;
END Test;

          CONST     max        = 31; (* ???? *)
                    Return     = CHR(10);
                    BackSpace  = CHR(8);
                    WrLine     = TRUE;
                    FehlFlag   = "Zeichenkette zu lang (TIO.RdRealDefault)";

          VAR       i,j,nc     : CARDINAL;
                    ch         : CHAR;
                    RealString : ARRAY [0..max] OF CHAR;
                    result     : ConvResults;
                    Fehler     : BOOLEAN;
BEGIN
      nc:=0;
      LOOP
        ReadNBytes(Input,ch,1,j);
        IF (ch = Return) THEN 
          IF (nc = 0) THEN
            X:=Y; EXIT;         (* Vorgabe-Wert zur"uckgeben ! *)
          ELSE
            RealString[nc]:=0C; (* Dateneingabe beendet !      *)
            StringToReal(RealString,X,result);
            IF (result # strAllRight) THEN
              WrStr("Fehler"); WrLn;
              (* L"osche Eingabefeld, neuer Versuch ! *) 
              FOR i:=nc TO 0 BY -1 DO WrChar(BackSpace); END;
              nc := 0;
            ELSE
              EXIT;
            END;
          END;
        ELSIF NOT Test(ch) THEN (* L"osche letztes Zeichen, neuer Versuch ! *) 
          WrChar(BackSpace);
        ELSE (* F"uge Zeichen ein *)
          RealString[nc] := ch; INC(nc);
        END;
        IF (nc >= max) THEN
          Fehler:=TRUE;
          Copy(Fehlerflag,FehlFlag);
          X:=MAX(LONGREAL);
          EXIT;
        END; 
      END; (* LOOP *)
      IF WrLine THEN WrLn; END;
END RdRealDefault;

BEGIN
      ExpAnz:=3;
      FFlag[strAllRight]    := 'In Ordnung !';
      FFlag[strOutOfRange]  := 'Wertebereichs"uberschreitung !';
      FFlag[strWrongFormat] := 'Falsches Format !';
      FFlag[strEmpty]       := 'Zeichenkette leer !';
END TIO.
