IMPLEMENTATION MODULE Timer; 

  (*------------------------------------------------------------------------*)
  (* Einige Routinen zur Zeiterfassung.                                     *)
  (* Some routines to measure timings.                                      *)
  (*------------------------------------------------------------------------*)
  (* Letzte Bearbeitung:                                                    *)
  (*                                                                        *)
  (*    11.94, MRi: Durchsicht, Anpassungen                                 *)
  (* 03.07.15, MRi: Einfuegen von  DeltaZeitHZr und UserTimeHZ              *)
  (* 07.07.16, MRi: Einfuegen von  GetUserTime aus Trudge, hier umgestellt  *)
  (*                auf Aufruf von UnixLib.time                             *)
  (*------------------------------------------------------------------------*)
  (* Modul ist stark systemabh"angig, dies ist die Implementation f"ur      *)
  (* Mocka (Linux und SUN), MtC, XDS M2 und GM2/Linux                       *)
  (*------------------------------------------------------------------------*)
  (* Offene Punkte                                                          *)
  (*                                                                        *)
  (* - GetUserTime testen                                                   *)
  (*------------------------------------------------------------------------*)
  (* Implementation : Michael Riedl                                         *)
  (* Licence        : GNU Lesser General Public License (LGPL)              *)
  (*------------------------------------------------------------------------*)

  (* $Id: Timer.mod,v 1.1 2015/09/13 18:26:40 mriedl Exp mriedl $ *)

             IMPORT SysClock;
FROM UnixLib IMPORT time,times,timeT;
             IMPORT Streams,SIO;

VAR   dz0, dzHZ0  : tms;
      init,initHZ : BOOLEAN;

PROCEDURE GetUserTime(VAR T : LONGCARD);

      VAR t : timeT;
BEGIN
      IF SysClock.CanGetClock() THEN
        t:=time(t);
        T := VAL(LONGCARD,t);
      END;
END GetUserTime;

PROCEDURE DeltaZeit() : CARDINAL;

          VAR dt : LONGCARD; 
              t  : tms; 
BEGIN
      IF init THEN
        times(t); 
        dt := (t.utime - dz0.utime) + (t.stime - dz0.stime); 
        dz0 := t;
        RETURN dt DIV SysClock.maxSecondParts;
      ELSE
        times(t); 
        dz0 := t;
        init:=TRUE;
        RETURN 0; 
      END;
END DeltaZeit;

PROCEDURE DeltaZeitHZ() : CARDINAL;

          VAR dt : LONGCARD; 
              t  : tms; 
BEGIN
      IF initHZ THEN
        times(t); 
        dt := (t.utime - dzHZ0.utime) + (t.stime - dzHZ0.stime); 
        dzHZ0 := t;
        RETURN dt;
      ELSE
        times(t); 
        dzHZ0 := t;
        initHZ:=TRUE;
        RETURN 0; 
      END;
END DeltaZeitHZ;

PROCEDURE DeltaZeitHZr() : REAL;

          VAR dt : LONGCARD; 
              t  : tms; 
BEGIN
      IF initHZ THEN
        times(t); 
        dt := (t.utime - dzHZ0.utime) + (t.stime - dzHZ0.stime); 
        dzHZ0 := t;
        RETURN VAL(REAL,dt) / VAL(REAL,SysClock.maxSecondParts);
      ELSE
        times(t); 
        dzHZ0 := t;
        initHZ:=TRUE;
        RETURN 0.0; 
      END;
END DeltaZeitHZr;

PROCEDURE InitUserTime(VAR t0 : tms);

BEGIN
      times(t0); 
END InitUserTime;

PROCEDURE UserTime(t0 : tms) : LONGCARD;

          VAR t  : tms;
              dt : CARDINAL;
BEGIN
      times(t);
      dt := (t.utime - t0.utime) + (t.stime - t0.stime); 
      RETURN dt DIV SysClock.maxSecondParts;
END UserTime;

PROCEDURE UserTimeHZr(t0 : tms) : REAL;

          VAR t  : tms;
              dt : CARDINAL;
BEGIN
      times(t);
      dt := (t.utime - t0.utime) + (t.stime - t0.stime); 
      RETURN VAL(REAL,dt) / VAL(REAL,SysClock.maxSecondParts);
END UserTimeHZr;

PROCEDURE WriteDate(VAR Aus  : Streams.Stream;
                    VAR Date : SysClock.DateTime); 

BEGIN 
      SIO.WrCardBN(Aus,Date.day,   2); SIO.WrChar(Aus,".");
      SIO.WrCardBN(Aus,Date.month, 2); SIO.WrChar(Aus,".");
      SIO.WrCardBN(Aus,Date.year,  4);
END WriteDate;

PROCEDURE WriteTime(VAR Aus  : Streams.Stream;
                    VAR Zeit : SysClock.DateTime); 

BEGIN 
      SIO.WrCardBN(Aus,Zeit.hour,  2); SIO.WrChar(Aus,":");
      SIO.WrCardBN(Aus,Zeit.minute,2); SIO.WrChar(Aus,":");
      SIO.WrCardBN(Aus,Zeit.second,2);
END WriteTime;

BEGIN
      init   := FALSE; times(dz0); 
      initHZ := FALSE; dzHZ0:=dz0;
END Timer.
